"""simple Newton method for task 1c"""
import numpy as np
import matplotlib.pyplot as plt


## initial definitions

# define y = [x1; x2; lambda]
# first order necessary conditions: F(y) = 0
def F(y):
    return np.array([
        -2*y[2]*y[0],
        1 - 2*y[2]*y[1],
        y[0]**2 + y[1]**2 - 1,
    ])


# Jacobian of root finding problem <=> KKT matrix
# We're using exact Newton's method
def J(y):
    return np.array([
        [-2*y[2], 0,      -2*y[0]],
        [0,      -2*y[2], -2*y[1]],
        [2*y[0],  2*y[1],  0],
    ])


# random initialization of y
# distribituted uniformly random in box around origin
y = 2*(np.random.rand(3,1) - 0.5) 

y = np.array([1, -1, 2])

tol = 1e-5;                 # convergence tolerance
maxit = 30;                 # maximum iterations

## main loop
for i in range(maxit):
    # check convergence
    if np.amax(np.abs(F(y))) < tol:
        print(f"solution found in {i-1} iterations")
        break
    # take full Newton step
    y = y - np.linalg.solve(J(y), F(y))
    print(f"y{i}={y}")


if i >= maxit:
    print("maximum iterations reached")


## plot result

theta = np.linspace(0, 2*np.pi, 100)
plt.figure(1)
plt.plot(np.sin(theta), np.cos(theta), "k", label=r"$g(x) = 0$")
plt.plot(y[0], y[1], "rx", label=r"$y^*$")
plt.axis("equal")
plt.legend()


plt.show()
